# Desactiva el bytecode de Python
import sys
sys.dont_write_bytecode = True

# Kafka
import kafka

# Librerías
import time
import atexit
import pandas as pd
from typing import List

# SQLAlchemy
import sqlalchemy as sa
import sqlalchemy.orm as orm

# Demonio
from demonio import Demonio

# Funciones
import general
import especie as Esp
import geolocalizacion as Geo

# Modelos
import modelos
from modelos import Registro, RegistroKafka, Especie
from modelos import SnibStateView, SnibMunView, SnibAgebView

# --------------------------- Ajustes del Consumidor --------------------------

# --------- Datos PostgreSQL ----------
PG_HOST       = 'base.datos.com'
PG_PUERTO     = 5433
PG_USUARIO    = 'usuario'
PG_CLAVE      = 'clave'
PG_BASE_DATOS = 'epi_puma'

# ------------ Datos Kafka ------------
KAFKA_BOOTSTRAP_SERVER = 'servidor.kafka:9092'
KAFKA_SCHEMA_REGISTRY  = 'http://servidor.kafka:8081'
KAFKA_GROUP_ID         = 'covid19_epi_puma'
KAFKA_TOPIC            = 'covid19'
KAFKA_PARTITION        = 0

# Tiempo sin recibir mensajes para considerar que hubo un cambio de día (segundos)
TIEMPO_PURGA_REGISTROS = 60

# Los registros se guardarán en paquetes de este tamaño (cantidad de registros)
LIMITE_CARGA_REGISTROS = 100000

# ----------------------------- Variables Globales ----------------------------

# Conexiones
CONEXION_POSTGRES: modelos.Conexion = None
CONEXION_KAFKA:    kafka.Conexion   = None

# Base de datos
MOTOR_SQL: sa.engine.Engine = None
SESION:    orm.Session      = None

# Variables globales
CAMBIO_DE_DIA            = False
BASE_ACTUALIZADA         = False
TRANSACCION_FINALIZADA   = False
TIEMPO_ANTERIOR: float   = None
DIA_ACTUAL: pd.Timestamp = None
INDICE_KAFKA: int        = None

# Variables globales de la carga de registros
INSERTADOS:   List[RegistroKafka] = []
ELIMINADOS:   List[RegistroKafka] = []
ACTUALIZADOS: List[RegistroKafka] = []
PENDIENTES:   List[RegistroKafka] = []

# -------------------------- Funciones de inserción ---------------------------

def insertar_registros(registros: pd.DataFrame):
	''' Inserta los registros procesados en la base de datos '''

	# Si no hay registros, no se hace nada
	if registros.empty:
		return

	# Campos para buscar una ubicación geográfica
	campos_geo = \
	[
		'longitud',
		'latitud'
	]

	# Campos para buscar una especie
	campos_esp = \
	[
		'grupo_edad',
		'sexo',
		'tipo',

		# Se obtienen a partir de 'tipo'
		'conjunto',
		'clasificacion'
	]

	# Función para cargar una ubicación geográfica desde la base de datos
	def cargar_geo(renglon: pd.Series):
		geo = Geo.cargar_o_crear_geolocalizacion(*[renglon[campo] for campo in campos_geo])

		# Se debe consultar un campo del modelo para obtener su '__dict__'
		geo.gid; return pd.Series(geo.__dict__)

	# Función para cargar una especie desde la base de datos
	def cargar_spe(renglon: pd.Series):
		spe = Esp.cargar_o_crear_especie(*[renglon[campo] for campo in campos_esp[:3]])

		# Se debe consultar un campo del modelo para obtener su '__dict__'
		spe.spid; return pd.Series(spe.__dict__)

	# Se cargan las ubicaciones geográficas en la base de datos
	geolocalizaciones = registros[campos_geo].drop_duplicates()
	geolocalizaciones = geolocalizaciones.apply(lambda renglon: cargar_geo(renglon), axis = 1)

	# Se unen las ubicaciones geográficas con los registros
	registros = registros.merge(geolocalizaciones, on = campos_geo)

	# Se cargan las especies en la base de datos
	especies = registros[campos_esp].drop_duplicates()
	especies = especies.apply(lambda renglon: cargar_spe(renglon), axis = 1)

	# Se unen las especies con los registros
	registros = registros.merge(especies, on = campos_esp)

	# Se actualizan las celdas de las especies
	especies.apply(lambda especie: Esp.actualizar_celdas_especie(registros, especie), axis = 1)

	# Se actualizan las celdas de las regiones
	especies.apply(lambda especie: Esp.actualizar_celdas_regiones(especie), axis = 1)

	# Se insertan los registros en la base de datos
	SESION.bulk_insert_mappings(Registro, registros.to_dict(orient = 'records'))
	SESION.flush()

def eliminar_registros(registros_kafka: List[RegistroKafka]):
	'''
		Elimina los registros de tipo kafka de la base de datos
		- Unicamente se toma en cuenta el campo 'ID_REGISTRO' para eliminar los registros
	'''

	# Si no hay registros, no se hace nada
	if len(registros_kafka) == 0:
		return

	# Se obtienen los ids de los registros de Kafka
	ids = [registro.ID_REGISTRO for registro in registros_kafka]
	ids_fallecidos = [f'{id}_f' for id in ids]

	# Se eliminan los registros de la base de datos
	SESION.query(Registro).filter(Registro.id.in_(ids)).delete()
	SESION.query(Registro).filter(Registro.id.in_(ids_fallecidos)).delete()
	SESION.flush()

def refrescar_datos_adicionales(conjunto = 'COVID-19'):
	''' Refresca las vistas materializadas y las celdas de las especies '''

	# Se refrescan las vistas materializadas 'snib_grid_statekm', 'snib_grid_munkm' y 'snib_grid_agebkm'
	SESION.execute(sa.text('REFRESH MATERIALIZED VIEW snib_grid_statekm'))
	SESION.execute(sa.text('REFRESH MATERIALIZED VIEW snib_grid_munkm'))
	SESION.execute(sa.text('REFRESH MATERIALIZED VIEW snib_grid_agebkm'))
	SESION.flush()

	# Se actualizan los arreglos 'cells_statekm_1' de la tabla 'sp_snib'
	subconsulta_state = SESION.query(SnibStateView.gridid_statekm) \
		.filter(SnibStateView.spid == Especie.spid).correlate(Especie).distinct().scalar_subquery()
	SESION.query(Especie).filter(Especie.conjunto == conjunto) \
		.update({Especie.cells_statekm_1: sa.func.array(subconsulta_state)}, synchronize_session = False)

	# Se actualizan los arreglos 'cells_munkm_1' de la tabla 'sp_snib'
	subconsulta_mun = SESION.query(SnibMunView.gridid_munkm) \
		.filter(SnibMunView.spid == Especie.spid).correlate(Especie).distinct().scalar_subquery()
	SESION.query(Especie).filter(Especie.conjunto == conjunto) \
		.update({Especie.cells_munkm_1: sa.func.array(subconsulta_mun)}, synchronize_session = False)

	# Se actualizan los arreglos 'cells_agebkm_1' de la tabla 'sp_snib'
	subconsulta_ageb = SESION.query(SnibAgebView.gridid_agebkm) \
		.filter(SnibAgebView.spid == Especie.spid).correlate(Especie).distinct().scalar_subquery()
	SESION.query(Especie).filter(Especie.conjunto == conjunto) \
		.update({Especie.cells_agebkm_1: sa.func.array(subconsulta_ageb)}, synchronize_session = False)

# ------------------------- Funciones del consumidor --------------------------

def cantidad_registros_pendientes():
	''' Regresa la cantidad de registros pendientes por cargar en la base de datos '''

	return len(INSERTADOS) + len(ELIMINADOS) + len(ACTUALIZADOS) + len(PENDIENTES)

def refrescar_dia_actual(fecha_registro: pd.Timestamp, purgar = False):
	'''
		Refresca la fecha actual según la fecha de un registro y determina si dicho registro
		corresponde a un dia diferente al actual, en cuyo caso se toma como un cambio de dia
	'''

	global DIA_ACTUAL, CAMBIO_DE_DIA

	# Si la fecha del registro es mayor a la fecha actual, se toma como un cambio de dia
	if fecha_registro > pd.Timestamp(DIA_ACTUAL):
		if purgar:
			print("Purgando registros pendientes...\n")
		DIA_ACTUAL = fecha_registro
		CAMBIO_DE_DIA = True

	# Si el dia actual es nulo, se asigna la fecha del registro
	elif DIA_ACTUAL is None:
		DIA_ACTUAL = fecha_registro

def enlistar_registro(registro: RegistroKafka):
	''' Enlista el registro en la lista correspondiente para su posterior procesamiento '''

	# Si se detecta un cambio de dia, se continúa con el siguiente ciclo y se guarda
	# el ultimo registro como pendiente para procesarlo en el siguiente ciclo
	if CAMBIO_DE_DIA:
		PENDIENTES.append(registro)
	elif registro.TYPE == 'INSERT':
		INSERTADOS.append(registro)
	elif registro.TYPE == 'DELETE':
		ELIMINADOS.append(registro)
	elif registro.TYPE == 'UPDATE':
		ACTUALIZADOS.append(registro)

def inicio():
	''' Función de inicialización del consumidor de Kafka '''

	global CONEXION_POSTGRES, CONEXION_KAFKA, MOTOR_SQL, SESION, TIEMPO_ANTERIOR

	# -------------------------- Finalización del script --------------------------

	def finalizar_proceso():
		''' Finaliza el script y cierra las conexiones a la base de datos y kafka '''

		print('\nFinalizando proceso...')
		CONEXION_KAFKA.cerrar()

		# Si la transacción no fue finalizada y hay algún registro pendiente por
		# cargar en la base de datos, se regresa el indice de kafka al inicio
		if not TRANSACCION_FINALIZADA or cantidad_registros_pendientes() > 0:
			CONEXION_KAFKA.regresar_indice_remoto()
			print('Indice de Kafka restaurado')

		# Se cierran las conexiones a postgres y kafka
		CONEXION_POSTGRES.cerrar()
		CONEXION_KAFKA.cerrar()

		print('Conexiones finalizadas')

	atexit.register(finalizar_proceso)

	# --------------------- Conexión a la base de PostgreSQL ----------------------

	print('\nConectando a la base de datos...')

	CONEXION_POSTGRES = modelos.ConexionPostgreSQL \
	(
		PG_HOST,
		PG_PUERTO,
		PG_USUARIO,
		PG_CLAVE,
		PG_BASE_DATOS
	)

	# Se usa el motor y la sesión como constantes globales
	MOTOR_SQL = CONEXION_POSTGRES.motor_sql
	SESION = CONEXION_POSTGRES.sesion

	# Se usa la conexión para el modelo de especie y geolocalización
	Esp.usar_conexion(CONEXION_POSTGRES)
	Geo.usar_conexion(CONEXION_POSTGRES)

	print('  Base de datos conectada\n')

	# --------------------------- Conexión a Kafka -------------------------------

	print('Iniciando consumidor de Kafka...')

	CONEXION_KAFKA = kafka.Conexion \
	(
		KAFKA_BOOTSTRAP_SERVER,
		KAFKA_GROUP_ID,
		KAFKA_SCHEMA_REGISTRY,
		KAFKA_TOPIC,
		KAFKA_PARTITION,
		RegistroKafka.desde_diccionario
	)

	print('  Consumidor de Kafka iniciado\n')

	# Se obtiene el tiempo inicial
	TIEMPO_ANTERIOR = time.time()

def ciclo():
	''' Ciclo principal del consumidor de Kafka '''

	global CAMBIO_DE_DIA, TRANSACCION_FINALIZADA, TIEMPO_ANTERIOR

	# Si se detecta un cambio de dia o se alcanza el limite de registros
	# cargados, se procesan los registros cargados hasta el momento
	if CAMBIO_DE_DIA or cantidad_registros_pendientes() >= LIMITE_CARGA_REGISTROS:

		# Se marca la transacción como no finalizada
		TRANSACCION_FINALIZADA = False

		print(f'Procesando {len(INSERTADOS)} registros a insertar...')
		print(f'Procesando {len(ELIMINADOS)} registros a eliminar...')
		print(f'Procesando {len(ACTUALIZADOS)} registros a actualizar...')

		# Se eliminan los registros de la lista de 'ELIMINADOS' y 'ACTUALIZADOS'
		eliminar_registros(ELIMINADOS)
		eliminar_registros(ACTUALIZADOS)

		print('Cargando los datos en memoria...')

		# Se inicializan las listas de registros
		insertados = general.inicializar_dataframe(INSERTADOS)
		actualizados = general.inicializar_dataframe(ACTUALIZADOS)

		print('Insertando registros en la base de datos...')

		# Se insertan los registros de la lista de 'INSERTADOS' y 'ACTUALIZADOS'
		insertar_registros(insertados)
		insertar_registros(actualizados)

		print('Persistiendo los cambios en la base de datos...')

		# Se hacen efectivos los cambios en la base de datos
		SESION.commit()

		# Se marca la transacción como finalizada
		TRANSACCION_FINALIZADA = True

		# Se recorre el indice inicial de kafka al indice del ultimo registro procesado
		CONEXION_KAFKA.avanzar_indice_local()

		print('    Base de datos actualizada!\n')

		# Se actualizan los datos restantes unicamente en cada cambio de dia
		if CAMBIO_DE_DIA:
			print("Cambio de dia detectado...\n")
			CAMBIO_DE_DIA = False

			# Se actualizan los datos adicionales
			refrescar_datos_adicionales()
			SESION.commit()

			print('    Datos adicionales actualizados!\n')

		# Se limpian las listas de registros
		INSERTADOS.clear()
		ELIMINADOS.clear()
		ACTUALIZADOS.clear()

	# Si no hay registros pendientes, se obtiene el siguiente registro
	if len(PENDIENTES) == 0:

		# Se obtiene el siguiente registro y el tiempo actual
		registro: RegistroKafka = CONEXION_KAFKA.obtener_registro()
		tiempo_actual = time.time()

		# Si el registro no es nulo, se refresca el tiempo anterior
		if registro is not None:
			TIEMPO_ANTERIOR = tiempo_actual

		# Si transcurre cierto tiempo sin recibir mensajes, se considera que hubo
		# un cambio de dia y se cargan los datos leídos hasta el momento
		elif tiempo_actual - TIEMPO_ANTERIOR >= TIEMPO_PURGA_REGISTROS:

			# Solo se refresca el dia actual si hay registros cargados en las listas
			if cantidad_registros_pendientes() > 0:
				refrescar_dia_actual(DIA_ACTUAL + pd.Timedelta(days = 1), purgar = True)

	# En caso contrario, se obtiene el siguiente registro de la lista de pendientes
	else:
		registro = PENDIENTES.pop(0)

	# Si el registro es nulo, se continúa con el siguiente ciclo
	if registro is None:
		return

	# Se obtiene la fecha del registro
	fecha_registro = registro.fecha_actualizacion

	# La fecha de actualización de los registros eliminados es la fecha del archivo
	# anterior, por lo que se incrementa en un dia para que coincida con el dia actual
	if registro.TYPE == 'DELETE':
		fecha_registro += pd.Timedelta(days = 1)

	# Se refresca el dia actual
	refrescar_dia_actual(fecha_registro)

	# Se enlista el registro en la lista correspondiente
	enlistar_registro(registro)

# Se muestra la interfaz de consola del demonio
Demonio(inicio, ciclo, 'consumidor.pid', 'consumidor.log').interfaz()