import pandas as pd
from typing import List
from pathlib import Path

# Modelos
from modelos import RegistroKafka

# Nivel geográfico
NIVEL_GEOGRAFICO = 'municipio'

# Ruta del archivo actual
RUTA = Path(__file__).resolve().parent

# Ruta del archivo de estados
RUTA_ESTADOS = RUTA / 'datos/estados.csv'

# Ruta del archivo de municipios
RUTA_MUNICIPIOS = RUTA / 'datos/municipios.csv'

# Diccionario de sexos
SEXOS = \
{
	1: 'M',
	2: 'H',
	99: 'NO ESPECIFICADO'
}

# Diccionario de nacionalidades
NACIONALIDADES = \
{
	1: 'MEXICANA',
	2: 'EXTRANJERA',
	99: 'NO ESPECIFICADO'
}

# Diccionario de estados
ESTADOS = pd.read_csv(RUTA_ESTADOS, dtype =
{
	'ID_ESTADO': int,
	'ESTADO': str,
	'LATITUD': float,
	'LONGITUD': float
}) \
.set_index('ID_ESTADO') \
.to_dict(orient = 'index')

# Diccionario de municipios
MUNICIPIOS = pd.read_csv(RUTA_MUNICIPIOS, dtype =
{
	'ID_MUNICIPIO': int,
	'ID_ESTADO': int,
	'MUNICIPIO': str,
	'LATITUD': float,
	'LONGITUD': float
}) \
.set_index \
([
	'ID_MUNICIPIO',
	'ID_ESTADO'
]) \
.to_dict(orient = 'index')

# Lista de grupos de edad
GRUPOS_EDAD = \
[
	'0-18',
	'19-30',
	'31-40',
	'41-49',
	'50-55',
	'56-60',
	'61-65',
	'66-70',
	'70-75',
	'76-120'
]

def inicializar_dataframe(registros: List[RegistroKafka]):
	''' Inicializa el dataframe con los valores de una lista de registros '''

	# Si no hay registros, se regresa un dataframe vacío
	if len(registros) == 0:
		return pd.DataFrame()

	# Se crea el dataframe con los registros y las columnas especificadas
	df = pd.DataFrame \
	(
		[registro.__dict__ for registro in registros],
		columns =
		[
			'ID_REGISTRO',
			'FECHA_ACTUALIZACION',
			'ENTIDAD_RES', 'MUNICIPIO_RES',
			'SEXO', 'FECHA_SINTOMAS', 'FECHA_DEF',
			'EDAD', 'CLASIFICACION_FINAL', 'NACIONALIDAD'
		]
	)

	# Se cambia el nombre de la columna 'ID_REGISTRO' a 'id'
	df.rename(columns={'ID_REGISTRO': 'id'}, inplace= True)

	# Se crea una nueva columna 'grupo_edad' con los grupos de edad
	grupos = [-1, 18, 30, 40, 49, 55, 60, 65, 70, 75, 1000000]
	df['grupo_edad'] = pd.cut(df['EDAD'], grupos, labels = GRUPOS_EDAD)
	del df['EDAD']

	# Se crea una nueva columna 'sexo'
	df['sexo'] = df['SEXO'].apply(lambda id: SEXOS[id])
	del df['SEXO']

	# Se crea una nueva columna 'conjunto' con el valor 'COVID-19'
	df['conjunto'] = 'COVID-19'

	# Función 'clasificación' que interpreta la columna de 'CLASIFICACION_FINAL' y 'FECHA_DEF'
	def clasificacion(valor):
		if valor in [1, 2, 3]:
			return 'CONFIRMADO'
		elif valor in [4, 5]:
			return 'INVIABLE'
		elif valor in [6]:
			return 'SOSPECHOSO'
		elif valor in [7]:
			return 'NEGATIVO'
		else:
			return 'DESCONOCIDO'

	# Se crea una nueva columna 'clasificación'
	df['clasificacion'] = df['CLASIFICACION_FINAL'].apply(clasificacion)
	del df['CLASIFICACION_FINAL']

	# Se crea una nueva columna 'tipo' y 'copia_tipo'
	df['tipo'] = df['conjunto'] + ' ' + df['clasificacion']
	df['copia_tipo'] = df['tipo']

	# Se crea una nueva columna nacionalidad
	df['nacionalidad'] = df['NACIONALIDAD'].apply(lambda id: NACIONALIDADES[id])
	del df['NACIONALIDAD']

	# Se crean las columnas de 'pais_geo', 'estado_geo', 'latitud' y 'longitud'
	df['pais_geo'] = 'México'

	if NIVEL_GEOGRAFICO == 'estado':
		df['estado_geo'] = df['ENTIDAD_RES'].apply(lambda id: ESTADOS[id]['ESTADO'])
		df['latitud'] = df['ENTIDAD_RES'].apply(lambda id: ESTADOS[id]['LATITUD'])
		df['longitud'] = df['ENTIDAD_RES'].apply(lambda id: ESTADOS[id]['LONGITUD'])

	elif NIVEL_GEOGRAFICO == 'municipio':

		# Se combinan las columnas de 'MUNICIPIO_RES' y 'ENTIDAD_RES' en una sola
		df['mun_ent'] = df['MUNICIPIO_RES'].astype(str) + ',' + df['ENTIDAD_RES'].astype(str)
		df['mun_ent'] = df['mun_ent'].apply(lambda ren: tuple([int(x) for x in ren.split(',')]))

		# Función para cargar cierto valor asociado a la columna de 'mun_ent'
		def valor_mun_ent(id: tuple, tipo: str):

			# Se intenta cargar la llave (ID_MUNICIPIO, ID_ESTADO)
			if id in MUNICIPIOS:
				return MUNICIPIOS[id][tipo]

			# Se intenta cargar la llave (999, ID_ESTADO)
			elif (999, id[1]) in MUNICIPIOS:
				return MUNICIPIOS[(999, id[1])][tipo]

			# Se regresa el valor de la llave (999, 99)
			else:
				return MUNICIPIOS[(999, 99)][tipo]

		# Se crean las columnas de 'estado_geo', 'latitud' y 'longitud'
		df['estado_geo'] = df['mun_ent'].apply(lambda id: valor_mun_ent(id, 'MUNICIPIO'))
		df['latitud'] = df['mun_ent'].apply(lambda id: valor_mun_ent(id, 'LATITUD'))
		df['longitud'] = df['mun_ent'].apply(lambda id: valor_mun_ent(id, 'LONGITUD'))

		# Se elimina la columna de 'mun_ent'
		del df['mun_ent']

	# Se eliminan las columnas de 'ENTIDAD_RES' y 'MUNICIPIO_RES'
	del df['ENTIDAD_RES']
	del df['MUNICIPIO_RES']

	# Se convierten los campos de fecha a datetime
	df['FECHA_ACTUALIZACION'] = pd.to_datetime(df['FECHA_ACTUALIZACION'])
	df['FECHA_SINTOMAS'] = pd.to_datetime(df['FECHA_SINTOMAS'])
	df['FECHA_DEF'] = pd.to_datetime(df['FECHA_DEF'], format= '%Y-%m-%d', errors= 'coerce')

	# Se cargan los valores de las columnas de fecha de ingreso
	df['dia_ingreso'] = df['FECHA_ACTUALIZACION'].dt.day
	df['mes_ingreso'] = df['FECHA_ACTUALIZACION'].dt.month
	df['anio_ingreso'] = df['FECHA_ACTUALIZACION'].dt.year
	del df['FECHA_ACTUALIZACION']

	# Se cargan los valores de las columnas de fecha de síntomas
	df['dia_sintomas_def'] = df['FECHA_SINTOMAS'].dt.day
	df['mes_sintomas_def'] = df['FECHA_SINTOMAS'].dt.month
	df['anio_sintomas_def'] = df['FECHA_SINTOMAS'].dt.year
	del df['FECHA_SINTOMAS']

	# Se crean nuevos registros para los casos fallecidos que son confirmados
	fallecidos = df[df['FECHA_DEF'].notnull() & (df['clasificacion'] == 'CONFIRMADO')].copy()
	fallecidos['clasificacion'] = 'FALLECIDO'

	# Se modifica el 'id' de los fallecidos agregando '_f'
	fallecidos['id'] = fallecidos['id'] + '_f'

	# Se actualizan las columnas de 'tipo' y 'copia_tipo'
	fallecidos['tipo'] = fallecidos['conjunto'] + ' ' + fallecidos['clasificacion']
	fallecidos['copia_tipo'] = fallecidos['tipo']

	# Se cargan los valores de las columnas de fecha de defunción
	fallecidos['dia_sintomas_def'] = fallecidos['FECHA_DEF'].dt.day
	fallecidos['mes_sintomas_def'] = fallecidos['FECHA_DEF'].dt.month
	fallecidos['anio_sintomas_def'] = fallecidos['FECHA_DEF'].dt.year
	del fallecidos['FECHA_DEF']
	del df['FECHA_DEF']

	# Se regresa el dataframe original concatenado con los fallecidos
	return pd.concat([df, fallecidos])